/* ndk - [ memory.c ]
 *
 * ndk maintains "page stacks" in order to manage memory in the physical
 * address space.  This code relies heavily on the paging hardware and
 * interface described in the pager source.
 *
 * If done correctly, it should be able to create an rZero implementation
 * of this memory management code, and the pager, to allow for
 * interchangeable methods of managing memory.  The two modules, however,
 * would probably have to be combined into one rZero module, due to
 * their reliance on one another.
 *
 * Please see:
 *   /doc/memoryManagement.kwd
 *   /doc/memoryMap.kwd
 *   /src/pager.{c,h}
 *
 * (c)2002 dcipher / neuraldk
 *           www.neuraldk.org
 */

#include "memory.h"

// global variables describing memory size
long totalMemory, totalPages;
long dmaMemorySize, appMemorySize;
long dmaMemoryPages, appMemoryPages;

// staring and ending pages which the kernel occupies
long kernelStart;
long kernelEnd;

pageStack appStack;  // memory available for apps (and kernel...)
pageStack dmaStack;  // memory below 16MB, used for DMA

void memoryInit(void) {
  /* currently ndk only reserves 4MB for DMA... and it
   * should be noted the OS itself is withen the first 4MB.
   * I might port this to an rZero module and allow it to
   * be run time configureable (through rZero module's
   * "init" routine).
   */

  printf("Usable memory: [Lower: %dkb, Upper %dkb]\n",
                             multibootInfo->mem_lower,
                             multibootInfo->mem_upper);

  // don't use mem_lower to calculate total memory... it'll return
  // 640k (ie, it doesn't include the bios).  Use 1024k (ie, 1MB)
  totalMemory = (multibootInfo->mem_upper + 1024) * 1024;
  totalPages  = totalMemory / PAGE_SIZE;

  // calculate the size (in bytes and pages) of the dma memory range
  dmaMemorySize  = 4*1024*1024;
  dmaMemoryPages = dmaMemorySize / PAGE_SIZE;

  // calculate the size (in bytes and pages) of the app memory range
  appMemorySize  = totalMemory - dmaMemorySize;
  appMemoryPages = totalPages - dmaMemoryPages;

  // output some info about or stacks
  //printf("bssEnd: 0x%x\n", bssEnd);
  printf("DMA Stack: %d pages\n", dmaMemoryPages);
  printf("App Stack: %d pages\n", appMemoryPages);

  // kernel is currently always fixed @ 1MB mark, but could be any length
  // Use bssEnd to calculate total space used by kernel, remember to round
  // kernelEnd _UP_ to the nearest page.
  kernelStart = (1024 * 1024) / PAGE_SIZE;
  kernelEnd = ((long)&bssEnd + PAGE_SIZE - ((long)&bssEnd & (PAGE_SIZE-1))) / PAGE_SIZE;
  kernelEnd++;

  // create both the dma and app stacks in memory above the kernel and
  // rZero modules
  memoryCreateStacks();
  //memoryAllocateKernel();

  return;
}

/* find a place to put the memory stack, above the kernel
 * and rZero modules
 * the kernel loads itself @1MB, with rZero modules afterwards
 * therefore, to find the end of the kernel, find the end of
 * the last rZero module
 */
void memoryCreateStacks(void) {
  module_t *mod;
  long addr;
  int i;

  // get the end address of the last module
  mod  = (module_t *) multibootInfo->mods_addr;
  addr = (long)mod[multibootInfo->mods_count-1].mod_end;

  // make sure it's at least dword aligned (rounding up!)
  addr += 4 - (addr & 3);
  printf("addr: 0x%8x\n", addr);

  /* The following calculations gave me much grief when
   * initially developing this code.  I had replaced addr
   * with dmaStack.base, or appStack.base when calculating
   * the used base pointer... although the two are exactly
   * equivalent (printf confirms), the calculation came
   * out wrong!  Just a word of warning...
   */

  // first put the dma stack, then the app stack
  printf("Creating DMA stack at: 0x%8x\n", addr);
  dmaStack.total           = dmaMemoryPages;
  dmaStack.avail.flags     = STACK_EXPAND_DOWN;
  dmaStack.avail.pointer   = 0; //dmaStack.total - 1;
  dmaStack.avail.base      = addr;
  dmaStack.used.flags      = STACK_EXPAND_UP;
  dmaStack.used.base       = (addr + 4 * dmaMemoryPages);
  dmaStack.used.pointer    = 0;
  //memoryStackSpecs(&dmaStack);

  addr = (long)dmaStack.used.base;  // NOTE: + 4 needed?
  printf("Creating Application stack at: 0x%8x\n", addr);
  appStack.total           = appMemoryPages;
  appStack.avail.flags     = STACK_EXPAND_DOWN;
  appStack.avail.pointer   = 0; //appStack.total - 1;
  appStack.avail.base      = addr;
  appStack.used.flags      = STACK_EXPAND_UP;
  appStack.used.base       = (addr + 4 * appMemoryPages);
  appStack.used.pointer    = 0;
  //memoryStackSpecs(&appStack);

  // fill the dma stack with all pages (unless they're already in
  // use, by the kernel)
  for(i = 0; i < dmaStack.total; i++) {
    if(!memoryPageUsedByKernel(i))
      stackPush(&dmaStack.avail, i * PAGE_SIZE);
  }

  // fill the app stack with all pages (unless in use by kernel)
  for(i = dmaMemoryPages; i < (appStack.total + dmaMemoryPages); i++) {
    if(!memoryPageUsedByKernel(i))
      stackPush(&appStack.avail, i * PAGE_SIZE);
  }

  // print some info 'bout kernel and stacks
  printf("Kernel start %d, end %d\n", kernelStart, kernelEnd);

  // what about bss of last module?
  // its dynamically allocated by this memory manager, and
  // linked in by the rZero code... it's all good! :)
}

/* Quick and dirty method to tell memoryCreateStacks wether a
 * certain page is already in use by the kernel (ie, it was used
 * before memory management was introduced).  If it's in use,
 * the page isn't included into the page stack.
 *
 * It should also be noted that it currently isn't included in
 * the "used" portion of the stack... there really isn't any
 * need for this, either... we aren't going to deallocate the
 * kernel!
 */
long memoryPageUsedByKernel(long page) {
  // if page is withen 0xA0000 and 0xF0000, inclusive, it's the
  // bios... don't allow this to be allocated! :)
  if(page >= 0xA0000/PAGE_SIZE && page <= 0xF0000/PAGE_SIZE)
    return 1;

  // if page is withen 0x90000 and 0x93000 (inc), it's being used for
  // system and vm86 page directory + first table, allocate it
  if(page >= 0x90000/PAGE_SIZE && page <= 0x93000/PAGE_SIZE)
    return 1;

  // if page is withen kernel address range, allocate it
  if(page >= kernelStart && page <= kernelEnd)
    return 1;

  // first two pages used by pmode idt + gdt and real mode idt
  if(page == 0 || page == 1) return 1;

  // we could also provide provisions to allocate the rZero modules
  // at this point... but this is already pretty slow... doing this
  // check on every single page in the system?  Eh...
  return 0;
}

/* Allocates one (4096) page, returning the physical address.
 * The function makes no attempt at mapping into the linear
 * address space!
 */
void *memoryAllocUnmappedPage(void) {
  long physAddr;

  if(appStack.avail.pointer == 0) {
    printf("Not enough physical memory!\n");
    return 0;
  }

  // place the physical address onto the used stack, and
  // return it
  physAddr = stackPop(&appStack.avail);
  stackPush(&appStack.used, physAddr);

  return (void *)physAddr;
}

void *memoryAllocPage(void) {
  long physAddr;
  long linAddr;

  physAddr = (long)memoryAllocUnmappedPage();
  //printf("Physical Addr: 0x%x\n", physAddr);
  linAddr = (long)memoryMapPage((void *)physAddr);
  //printf("Linear Addr:   0x%x\n", linAddr);

  return (void *)linAddr;
}

void memoryFreePage(void *linAddr) {
  long *pages = (long *)0xffc00000;
  long  physAddr;
  pageStack *pStack;
  long  i = 0;
  long  pTable, pTableIndex;
  unsigned long pageOffset;

  // convert the linear address to a physical one
  // NOTE: because of the "and" mask, this function will
  // deallocate addresses that aren't page aligned.
  // ie, 0x1005 will remove 0x1000
  linAddr  = (void *)((long)linAddr & 0xfffff000);
  physAddr = (long)pagerLinToPhys(linAddr);
  printf("removing physical address 0x%8x\n", physAddr);

  // select the correct page stack, based on where the physical
  // address resides
  if((long)physAddr < TOP_OF_DMA_STACK) pStack = &dmaStack;
  else pStack = &appStack;

  // find the physical address in the used stack
  while(stackPeek(&pStack->used, i) != physAddr) {
    i++;
    if(i == pStack->used.pointer) {
      printf("Cannot deallocate 0x%8x; Not allocated!\n", physAddr);
      return;
    }
  }

  printf("Copying from used to available stack\n");
  // copy address from used stack, to available stack
  // and "increment" pointers accordingly.
  stackRemove(&pStack->used, i, 1);
  stackPush(&pStack->avail, physAddr);


  // now remove page from page directory (include in pager.c?!)
  pageOffset = (unsigned long)linAddr >> 12;
  printf("Removing from page directory (%d)\n", pageOffset);

  printf("Invalidating TLB (0x%8x)\n", (unsigned long)linAddr);
  invalidatePage( (unsigned long)linAddr );

  pages[(unsigned long)pageOffset] = 0;

  return;
}

/*
void *memoryAllocatePages(long numPages) {
  void *linStartAddr;
  long  physStartAddr;
  long  i, pointer;

  // check if there's enough pages available
  if(appStack.pointer < numPages) {
    printf("Not enough physical memory!\n");
    return 0;
  } else {
    // remove these pages from the stack right away, so any other
    // attempt to call this function won't allocate the same
    // range of memory... or should I loop through this one at a time... (yes)
    pointer = appStack.pointer;
    appStack.pointer -= numPages;
  }

  // keep track of the starting page in physical and linear address space
  physStartAddr = appStack.base[pointer--];
  printf("First page, physical: 0x%8x ", physStartAddr);

  // find a chunk in linear address space that'll hold the newly allocated
  // memory
  linStartAddr = (void *)memoryFindLinearChunk(numPages);
  if(linStartAddr = 0x0) {
    printf("Not enough linear memory!\n");
    return 0;
  }
  //linStartAddr = (void *)pagerMapPage(physStartAddr);
  printf("linear 0x%8x\n", linStartAddr);

  // map the rest of the pages into the apps linear address space
  for(i = 1; i < numPages; i++) {
    printf("Allocating page #%d\n", i);
    pagerMapPage(appStack.base[pointer--]);
  }

  printf("Marking pages as used\n");
  // now mark these pages into the "used" stack
  if(numPages > 1) {
    printf("  more than one\n");
    *(long *)(appStack.used.base - (appStack.used.pointer++)) = numPages;
    *(long *)(appStack.used.base - (appStack.used.pointer++)) = physStartAddr &
                                                            (MULTI_PAGE_ENTRY);
  }
  else {
    printf("  just one\n");
    *(long *)(appStack.used.base - (appStack.used.pointer++)) = physStartAddr;
  }
  printf("returning\n");
  return linStartAddr;
}
*/

/*
void memoryFreePages(void *addr) {
  long  pageNum = 0;
  long  numPages;
  long  pointer;
  long *pages;
  long  physAddr;
  pageStack *pStack;

  // Because the page directory is mapped into itself, as the last
  // page table, all physical pages in the linear address space are
  // included in the top 4MB...
  //
  pages = (long *)0xffc00000;

  // addr will be in linear address space... convert it to physical
  while(pages[pageNum] != (long)addr) pageNum++;
  physAddr = pages[pageNum];

  if((long)physAddr < TOP_OF_DMA_STACK) pStack = &dmaStack;
  else pStack = &appStack;

  // now search for this physical address in the used page stack
  pointer = pStack->used.pointer;
  while( (pStack->used.base[pointer] & 0xfffff000) != physAddr) {
    if(pStack->used.base[pointer] & MULTI_PAGE_ENTRY) pointer -= 2;
    else pointer -= 1;
  }

  // how many pages are associated with this deallocation?
  if(pStack->used.base[pointer] & MULTI_PAGE_ENTRY)
    numPages = pStack->used.base[pointer-1];
  else numPages = 1;

  // remove pages from the used stack...
  // copy top of stack overtop of, soon-to-be, deallocated page
  pStack->used.base[pointer] = pStack->used.base[pStack->used.pointer];
  pStack->used.pointer--;

  // if this is a multi-page deallocation, copy over the page count too
  if(numPages > 1) {
    pStack->used.base[pointer-1] = pStack->used.base[pStack->used.pointer];
    pStack->used.pointer--;
  }

  // and, finally, remove these pages from the tasks paging directory
  while(numPages--)
    pages[pageNum++] = 0;
}
*/

/* Maps physAddr into the linear address space where-ever there
 * is an available page (should be in pager.c!?)
 */
void *memoryMapPage(void *physAddr) {
  long *pages = (long *)0xffc00000;
  long  pageNum = ALLOC_OFFSET;
  long  pTableNum;

  printf("memoryMapPage: 0x%8x\n", physAddr);
  pTableNum = pageNum / 1024;
  pageNum &= 1023;
  while(1) {
    // if the current page table isn't allocated, create it
    if(!pagerTableExists(pTableNum))
      pagerCreateTable(pTableNum);

    // if this page is empty, use it, and return the
    // linear address
    // NOTE: pages[pageNum] not correct... pageNum = 0...1023
    if(pages[(pTableNum*1024)+pageNum] == 0) {
      pages[(pTableNum*1024)+pageNum] = (long) physAddr  | PAGE_PRESENT
                                                         | PAGE_READ_WRITE
                                                         | PAGE_P3_ACCESS;
      return (void *)((pTableNum*1024)*4096)+(pageNum * 4096);
    }

    // incrememnt to the next page
    pageNum++;

    // if we've gone through 1024 pages, go to the next table
    if(pageNum == 1024) {
      pTableNum++;
      pageNum = 0;
    }

    // out of memory!
    if(pTableNum == 1024) return 0;
  }
}

/*
void *memoryFindLinearChunk(long size) {
  void *linearAddr;
  long *pages = 0xffc00000;
  //long *pgDir = 0xfffff000;
  long  pTableNum;
  long  pageNum;
  // start allocating pages at the 2GB mark (linear)
  long  startAtPage = ALLOC_OFFSET;//524288;
  long  remainingPages = size;

  pageNum = startAtPage;
  remainingPages = size;
  pTableNum = pageNum / 1024;
  while(remainingPages) {
    if(!pagerTableExists(pTableNum))
      pagerCreateTable(pTableNum);

    // is the next page empty?
    if(pages[pageNum] == 0) {
      remainingPages--;
    } else {
      remainingPages = size;
    }

    // incrememnt to the next page
    pageNum++;

    // if we've gone through 1024 pages, go to the next table
    if(pageNum == 1024) {
      pTableNum++;
      pageNum = 0;
    }
  }

  return linearAddr;
}
*/

/* Allocate only the kernel data, code and bss sections.  The rZero
 * modules will be allocated via separate calls to the memory manager,
 * so that they may be removed from memory independantly from the
 * kernel (such as the splashScreen, which serves no purpose in memory
 * past boot up).
 */
/*
void memoryAllocateKernel(void) {
  // allocate the kernel itself
  memoryAllocateUsedPages(kernelStart, kernelEnd - kernelStart);
  // allocate the pmode idt + gdt, and real mode idt
  memoryAllocateUsedPages(0, 2);
  // allocate the kernel + vm86 page directory + table
  memoryAllocateUsedPages(0x90000/PAGE_SIZE, 4);
}
*/

/* This function is used the allocate pages that have been used by the
 * kernel, prior to memory management being introduced
 */
/*
void memoryAllocateUsedPages(long startPage, long totalPages) {
  pageStack *ps;
  long i;
  long startAddr = startPage * PAGE_SIZE;

  // allocate from the correct stack (problems arise
  // if the allocation spans tables, but it never should
  // at this point... the kernel should pretty much
  // always reside withen the dmaStack anyway)
  if(startPage < dmaStack.total) pageStack = &dmaStack;
  else pageStack = &appStack;

  // find the right page
  for(i = 0; i < ps->pointer; i++)
    if(startAddr == ps->base[i]) {
      // remove these pages from the stack, copying the
      // other address upwards
      memcpy(&ps->base[i], &ps->base[i] + (totalPages*4),
             &ps->base[pointer] - (startPage + totalPages);
    }
}
*/

void memoryStackSpecs(pageStack *pStack) {
  printf("  Total:     %d\n"
         "  pointer:   %d\n"
	 "  base:      0x%8x\n"
	 "  Used:\n"
	 "    pointer: %d\n"
	 "    base:    0x%8x\n",
        pStack->total,
        pStack->avail.pointer,
        pStack->avail.base,
        pStack->used.pointer,
        pStack->used.base
  );
}
