/* ndk - [ pager.c ]
 *
 * Contains code to initialize the pager, and
 * to convert to and from linear and physical
 * address spaces.
 *
 * Please see:
 *   /src/memory.c
 *
 * (c)2002 dcipher / neuraldk
 *           www.neuraldk.org
 */
#include "pager.h"

pageDirectory *sysPageDir;

void pagerInit(void) {
  pageTable *pTable;
  long i, addr;

  // find 4kb aligned address for page directory
  sysPageDir = (pageDirectory *)0x00090000;

  // find 4kb aligned address for 1 page table
  pTable = (pageTable *)0x00091000;

  // assign 1st page directory entry to page table
  sysPageDir->pageTable[0] = (long)pTable | PAGE_PRESENT
                                          | PAGE_READ_WRITE
                                          | PAGE_P3_ACCESS;
  // assign the last page table to the page directory!
  // This will aid in memory management (doc/memoryManagement.kwd)
  sysPageDir->pageTable[1023] = (long)sysPageDir | PAGE_PRESENT
                                                 | PAGE_READ_WRITE
                                                 | PAGE_P3_ACCESS;

  // clear the rest of the page directory
  for(i = 1; i < 1023; i++)
    sysPageDir->pageTable[i] = 0;

  // assign the page table simple linear address for the
  // first 4MBs of memory (later this should be changed to
  // map only used memory)
  //addr = 0;
  for(i = 0; i < 1024; i++) {
    pTable->pageAddr[i] = (i * 4096) | PAGE_PRESENT
                                     | PAGE_READ_WRITE
                                     | PAGE_P3_ACCESS;
    //printf("pTable[%d] = 0x%8x\n", i, pTable->pageAddr[i]);
  }

  // load CR3 with the address of our page directory
  printf("setting CR3: (0x%8x)\n", sysPageDir);
  setCR3( ((getCR3() & 0xfff) | (long)sysPageDir) );
  printf("CR3: 0x%x\n", getCR3() );

  // and enable the paging bit in CR0
  printf("setting CR0: (0x%8x)\n", getCR0() | CR0_PAGING_ENABLE);
  setCR0( (getCR0() | CR0_PAGING_ENABLE) );
  printf("CR0: 0x%x\n", getCR0() );
}

long pagerTableExists(unsigned long tableNum) {
  long *pgDir = (long *)0xfffff000;

  return pgDir[(unsigned long)tableNum];
}

void pagerCreateTable(unsigned long tableNum) {
  long *pageDirectory = (long *)0xfffff000;
  long *pages = (long *)0xffc00000;
  long *pageTable;
  long  i;

  printf("pager: creating new table #%d\n", tableNum);
  // get a page from the kernel
  pageTable = (long *)memoryAllocUnmappedPage();
  printf("received page 0x%8x\n", pageTable);

  printf("Adding to directory\n");
  // and assign it to our page directory (must do before we write to it)
  pageDirectory[tableNum] = (long)pageTable  | PAGE_PRESENT
                                             | PAGE_READ_WRITE
                                             | PAGE_P3_ACCESS;;

  // zero it out
  printf("And clearing\n");
  for(i = 0; i < 1024; i++) pages[tableNum*1024+i] = 0;//pageTable[i] = 0;
}

long pagerMapPage(unsigned long physAddr) {
  long *pages = (long *)0xffc00000;
  long  allocPageOffset = (4*1024*1024)/4096;
  long  pointer;

  // simple way to find available slot
  pointer = allocPageOffset;
  while(pages[pointer++] != 0) ;

  // put physical page into linear space
  pages[pointer] = physAddr | PAGE_PRESENT
                            | PAGE_READ_WRITE
                            | PAGE_P3_ACCESS;
  return pointer*1024;
}

void *pagerLinToPhys(void *linearAddr) {
  long *pages = (long *)0xffc00000;
  long  linAddr = (long)linearAddr;
  long  physAddr;
  long  bottom4k;
  unsigned long pageTableNum;
  unsigned long pageIndex;

  // save the bottom 4k, and remove it from the linear address
  bottom4k = linAddr & 4095;
  linAddr -= bottom4k;

  // debugging info...
  printf("page aligned linAddr: 0x%8x\n", linAddr);

  // calculate which page table to look into
  pageTableNum   = (unsigned long)linAddr / (4096*1024);
  //printf("table %d, index %d\n", pageTableNum, pageTableIndex);
  // Check if this page table exists!
  if(!pagerTableExists(pageTableNum)) {
    printf("Page table doesn't exist for linear address 0x%8x\n", linAddr);
    return 0;
  }

  // calculate the offset into the page tables, and get our physical address
  pageIndex = (unsigned long)linAddr >> 12;
  physAddr = pages[(unsigned long)pageIndex] & 0xfffff000;

  // add the bottom 4k
  physAddr += bottom4k;

  return (void *)physAddr;
}
