; ndk - [ start.asm ]
;
; The assembly language stub which defines its
; sections to the multiboot loader, and sets up
; the environment so that a transfer to the C
; source code can be made.
;
; Please see:
;   /src/ndk.c
;
; (c)2002 dcipher / neuraldk
;           www.neuraldk.org

section .text
bits 32

; multi boot header defines
MBOOT_PAGE_ALIGN   equ 1 << 0
MBOOT_MEM_INFO     equ 1 << 1
MBOOT_AOUT_KLUDGE  equ 1 << 16

MBOOT_MAGIC equ 0x1BADB002
MBOOT_FLAGS equ MBOOT_PAGE_ALIGN | MBOOT_MEM_INFO | MBOOT_AOUT_KLUDGE
CHECKSUM    equ -(MBOOT_MAGIC + MBOOT_FLAGS)

STACK_SIZE  equ 0x1000

; defined in the linker script
extern textEnd
extern dataEnd
extern bssEnd
extern ndkStart
extern mb_main

global  start
global _start

entry:
    jmp start

    ; The Multiboot header
align 4, db 0
mBootHeader:
    dd MBOOT_MAGIC
    dd MBOOT_FLAGS
    dd CHECKSUM
    ; fields used if MBOOT_AOUT_KLUDGE is set in
    ; MBOOT_HEADER_FLAGS
    dd mBootHeader                     ; these are PHYSICAL addresses
    dd entry                           ; start of kernel .text (code) section
    dd dataEnd                         ; end of kernel .data section
    dd bssEnd                          ; end of kernel BSS
    dd entry                           ; kernel entry point (initial EIP)

 start:
_start:
    ; can't include this here... corrupts eax, ebx from multiboot
;    mov edi, 0xB8000
;    mov esi, string
;    mov ah, 0x0F
;  .displaying:
;    lodsb
;    stosw
;    or al, al
;    jnz .displaying

    ; make sure we're going in the right direction!
    cld

    ; copy the current real mode idt + bios info
    ; to a safe page (it's about to be overwritten)
    xor esi, esi
    mov edi, 0x00001000
    mov ecx, 0x1000
    rep movsb

    ; clear the idt + gdt @ 0x0
    push eax
    xor eax, eax
    xor esi, esi
    mov ecx, 0x800 + 0x800
    rep stosb
    pop eax

    ; setup a bare bones GDT
    mov esi, my_gdt
    mov edi, 0x0800
    mov ecx, 8 * 7
    rep movsb

    lgdt    [pGDT]                  ; load the GDT
    lidt    [pIDT]                  ; load the IDT
    lldt    [pLDT]                  ; load the LDT (0)

    mov     dx, 0x08        ; 0x08 - kernel data segment
    mov     ds, dx
    mov     es, dx
    mov     fs, dx
    mov     gs, dx
    mov     dx, 0x18        ; 0x18 - kernel stack segment
    mov     ss, dx

    mov    esp, (stack + STACK_SIZE)

    ; push the multiboot info structure, and magic
    push ebx
    push eax

    ; load cs with new selector
    jmp 0x10:new_gdt

  new_gdt:
    ; time for some C!

    ; enable the alignment check bit, for fun :)
    ; Personally, I think every single program should be
    ; properly aligned, but too many people don't know about
    ; the performance penalties!  With this bit enabled, the
    ; processor will interrupt on any unaligned memory
    ; access (assuming eflags also contains the same bit)
    ; In time, I'll probably just contain a running count
    ; of unaligned access per task... just for show (it'll
    ; slow unaligned programs even more... but... tough!)
    mov eax, cr0
    or eax, 262144
    mov cr0, eax

    call ndkStart
    ;call mb_main
    jmp short $

section .data
string          db "ndk is alive!", 0

pIDT            dw 800h         ; limit of 256 IDT slots
                dd 00000000h    ; starting at 0

pGDT            dw 800h         ; 256 GDT slots
                dd 00000800h    ; starting at 800h (after IDT)

pLDT            dw 0            ; no LDT at this time, but must still load
                dd 0            ; with someone (else multitasking might fail)

my_gdt:
        ; Null descriptor
        ;   base : 0x00000000
        ;   limit: 0x00000000 ( 0.0 MB)
        dd 0
        dd 0

        ; 0x08 descriptor - Kernel data segment
        ;   base : 0x00000000
        ;   limit: 0xfffff pages (4 GB)
        ;   DPL  : 0
        ;   32 bit, present, system, expand-up, writeable
        dd 0x0000ffff
        dd 0x00cf9200

        ; 0x10 descriptor - Kernel code segment
        ;   base : 0x00000000
        ;   limit: 0xfffff (4 GB)
        ;   DPL  : 0
        ;   32 bit, present, system, non-conforming, readable
        dd 0x0000ffff
        dd 0x00cf9A00

        ; 0x18 descriptor - Kernel stack segment
        ;   base : 0x00000000 //0x00080000
        ;   limit: 0xfffff (4 GB)
        ;   DPL  : 0
        ;   32 bit, present, system, expand-up, writeable
        dd 0x0000ffff
        dd 0x00cf9200
        ;dd 0x00cb9200

        ; ----

        ; 0x20 descriptor - Task data segment
        ;   base : 0x00000000
        ;   limit: 0xfffff pages (4 GB)
        ;   DPL  : 3
        ;   32 bit, present, system, expand-up, writeable
        dd 0x0000ffff
        dd 0x00cff200

        ; 0x28 descriptor - Task code segment
        ;   base : 0x00000000
        ;   limit: 0xfffff (4 GB)
        ;   DPL  : 3
        ;   32 bit, present, system, non-conforming, readable
        dd 0x0000ffff
        dd 0x00cffA00

        ; 0x30 descriptor - Task stack segment
        ;   base : 0x00000000 //0x00080000
        ;   limit: 0xfffff (4 GB)
        ;   DPL  : 3
        ;   32 bit, present, system, expand-up, writeable
        dd 0x0000ffff
        dd 0x00cff200

section .bss
    align 4, db 0
    common stack 0x1000
    resb 0x4000

